/*****************************************************************************/

/*
 *      devtree.c  --  USB device tree.
 *
 *      Copyright (C) 1999 Thomas Sailer, sailer@ife.ee.ethz.ch
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Please note that the GPL allows you to use the driver, NOT the radio.
 *  In order to use the radio, you need a license from the communications
 *  authority of your country.
 *
 */

/*****************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/dir.h>
#include <fcntl.h>

#include "devtree.h"

/* ---------------------------------------------------------------------- */

LIST_HEAD(usbbuslist);

/* ---------------------------------------------------------------------- */

static void freedev(struct usbdevnode *dev)
{
	free(dev);
}

static void freebus(struct usbbusnode *bus)
{
	free(bus);
}

/* ---------------------------------------------------------------------- */

static void markdel(struct list_head *list)
{
	struct usbdevnode *dev;
	struct list_head *list2;

	for (list2 = list->next; list2 != list; list2 = list2->next) {
		dev = list_entry(list2, struct usbdevnode, list);
		dev->flags |= USBFLG_DELETED;
		markdel(&dev->childlist);
	}
}

void devtree_markdeleted(void)
{
	struct usbbusnode *bus;
	struct list_head *list;
	
	for(list = usbbuslist.next; list != &usbbuslist; list = list->next) {
		bus = list_entry(list, struct usbbusnode, list);
		markdel(&bus->childlist);
	}
}

struct usbbusnode *devtree_findbus(unsigned int busn)
{
	struct usbbusnode *bus;
	struct list_head *list;
	
	for(list = usbbuslist.next; list != &usbbuslist; list = list->next) {
		bus = list_entry(list, struct usbbusnode, list);
		if (bus->busnum == busn)
			return bus;
	}
	return NULL;
}

static struct usbdevnode *findsubdevice(struct list_head *list, unsigned int devn)
{
	struct usbdevnode *dev, *dev2;
	struct list_head *list2;

	for (list2 = list->next; list2 != list; list2 = list2->next) {
		dev = list_entry(list2, struct usbdevnode, list);
		if (dev->devnum == devn)
			return dev;
		dev2 = findsubdevice(&dev->childlist, devn);
		if (dev2)
			return dev2;
	}
	return NULL;
}

struct usbdevnode *devtree_finddevice(struct usbbusnode *bus, unsigned int devn)
{
	return findsubdevice(&bus->childlist, devn);
}

/* ---------------------------------------------------------------------- */

int findtree(char *path, int level, int parentdevnum)
{
	int busnum, maxchild, productid, vendorid, devnum;
	int fp, i;
	char buffer[1024];
	char buf[100];
	char *cp;

	/* Read bus number */
	snprintf(buffer, sizeof(buffer), "%s/busnum", path);

	if ((fp = open(buffer, O_RDONLY)) == -1)
		return(0);
	read(fp, buf, sizeof(buf) - 1);
	busnum = strtoul(buf, NULL, 0);
	close(fp);

	/* Read max number of children. If >0 then it is a hub, so need to go
	 * through it again.
	 */
	snprintf(buffer, sizeof(buffer), "%s/maxchild", path);
	if ((fp = open(buffer, O_RDONLY)) == -1)
		return(0);
	read(fp, buf, 4);
	maxchild = strtoul(buf, NULL, 0);
	close(fp);

	/* Read device number */
	snprintf(buffer, sizeof(buffer), "%s/devnum", path);
	if ((fp = open(buffer, O_RDONLY)) == -1)
		return(0);
	read(fp, buf, 4);
	devnum = strtoul(buf, NULL, 0);
	close(fp);

	/* Read product id of device */
	snprintf(buffer, sizeof(buffer), "%s/idProduct", path);
	if ((fp = open(buffer, O_RDONLY)) == -1)
		return(0);
	read(fp, buf, 8);
	productid = strtoul(buf, NULL, 16);
	close(fp);

	/* Read vendor id of device */
	snprintf(buffer, sizeof(buffer), "%s/idVendor", path);
	if ((fp = open(buffer, O_RDONLY)) == -1)
		return(0);
	read(fp, buf, 8);
	vendorid = strtoul(buf, NULL, 16);
	close(fp);

	/* Add data to usb list */
	updateusblist(devnum, busnum, parentdevnum, level, vendorid, productid);

	/* Traverse through the connected devices */
	for (i = 1; i < maxchild; i++) {
		if (level >= 1) {
			cp = strrchr(path, '/');
			snprintf(buffer, sizeof(buffer), "%s/%s.%i", path, cp, i);
		} else
			snprintf(buffer, sizeof(buffer), "%s/%i-%i", path, busnum, i);
		findtree(buffer, level + 1, devnum);
	}
	return(1);
}

/*------------------------------------------------------------------*/

void updateusblist(unsigned int devnum, unsigned int busnum, unsigned int parentdevnum, unsigned int level, unsigned int vendor, unsigned int prodid)
{
	struct usbbusnode *bus;
	struct usbdevnode *dev, *dev2;

	devtree_markdeleted();
	if (!(bus = devtree_findbus(busnum))) {
		if (!(bus = malloc(sizeof(struct usbbusnode))))
			lprintf(0, "Out of memory\n");
		bus->busnum = busnum;
		bus->flags = USBFLG_NEW;
		INIT_LIST_HEAD(&bus->childlist);
		list_add_tail(&bus->list, &usbbuslist);
	} else
		bus->flags &= ~USBFLG_DELETED;
	if (!(dev = devtree_finddevice(bus, devnum)) || dev->vendorid != vendor || dev->productid != prodid) {
		if (!(dev = malloc(sizeof(struct usbdevnode))))
			lprintf(0, "Out of memory\n");
		dev->devnum = devnum;
		dev->flags = USBFLG_NEW;
		dev->bus = bus;
		dev->vendorid = vendor;
		dev->productid = prodid;
		INIT_LIST_HEAD(&dev->childlist);
		if (level == 0 && parentdevnum == 0) {
			list_add_tail(&dev->list, &bus->childlist);
			dev->parent = NULL;
		} else {
			if (!(dev2 = devtree_finddevice(bus, parentdevnum)))
				lprintf(0, "Bus %d Device %d Parent Device %d not found\n", busnum, devnum, parentdevnum);
			dev->parent = dev2;
			list_add_tail(&dev->list, &dev2->childlist);
		}
	} else
		dev->flags &= ~USBFLG_DELETED;
}

/* ---------------------------------------------------------------------- */

static void deletetree(struct list_head *list, unsigned int force)
{
	struct usbdevnode *dev;
	struct list_head *list2;

	for (list2 = list->next; list2 != list;) {
		dev = list_entry(list2, struct usbdevnode, list);
		list2 = list2->next;
		deletetree(&dev->childlist, force || dev->flags & USBFLG_DELETED);
		if (!force && !(dev->flags & USBFLG_DELETED))
			continue;
		list_del(&dev->list);
		INIT_LIST_HEAD(&dev->list);
		devtree_devdisconnect(dev);
		freedev(dev);
	}
}

static void newtree(struct list_head *list)
{
	struct usbdevnode *dev;
	struct list_head *list2;

	for (list2 = list->next; list2 != list; list2 = list2->next) {
		dev = list_entry(list2, struct usbdevnode, list);
		if (dev->flags & USBFLG_NEW)
			devtree_devconnect(dev);
		dev->flags &= ~USBFLG_NEW;
		newtree(&dev->childlist);
	}
}

void devtree_processchanges(void)
{
	struct list_head *list;
	struct usbbusnode *bus;

	for (list = usbbuslist.next; list != &usbbuslist;) {
		bus = list_entry(list, struct usbbusnode, list);
		list = list->next;
		deletetree(&bus->childlist, bus->flags & USBFLG_DELETED);
		if (!(bus->flags & USBFLG_DELETED))
			continue;
		list_del(&bus->list);
		INIT_LIST_HEAD(&bus->list);
		devtree_busdisconnect(bus);
		freebus(bus);
	}
	for (list = usbbuslist.next; list != &usbbuslist; list = list->next) {
		bus = list_entry(list, struct usbbusnode, list);
		if (bus->flags & USBFLG_NEW)
			devtree_busconnect(bus);
		bus->flags &= ~USBFLG_NEW;
		newtree(&bus->childlist);
	}
}

/* ---------------------------------------------------------------------- */

static void dumpdevlist(struct list_head *list, unsigned int level, unsigned int mask)
{
	struct usbdevnode *dev;
	struct list_head *list2;
	char buf[512];
	char *cp;
	unsigned int i;

	for (list2 = list->next; list2 != list; ) {
		dev = list_entry(list2, struct usbdevnode, list);
		list2 = list2->next;
		for (cp = buf, i = 0; i < level; i++) {
			*cp++ = (mask & (1 << i)) ? '|' : ' ';
			*cp++ = ' ';
		}
		if (list2 != list) {
			mask |= (1 << level);
			*cp++ = '|';
		} else {
			mask &= ~(1 << level);
			*cp++ = '`';
		}
		*cp++ = '-';
		snprintf(cp, buf + sizeof(buf) - cp, "Dev# %3d Vendor 0x%04x Product 0x%04x",
			 dev->devnum, dev->vendorid, dev->productid);
		lprintf(1, "%s\n", buf);
		dumpdevlist(&dev->childlist, level+1, mask);
	}
}

void devtree_dump(void)
{
	struct list_head *list;
	struct usbbusnode *bus;

	for (list = usbbuslist.next; list != &usbbuslist; list = list->next) {
		bus = list_entry(list, struct usbbusnode, list);
		lprintf(1, "Bus# %2d\n", bus->busnum);
		dumpdevlist(&bus->childlist, 0, 0);
	}
}
